@@ -1303,14 +1303,14 @@ https://arxiv.org/abs/2208.04720v2
     return ( decomposition )
 
 def generate_clustering_labels ( distm:np.array , cmd:str='min' , labels:list[str]=None ,
-                                 bExtreme:bool=False , n_clusters:int = None) -> tuple :
+                                 bExtreme:bool=False , n_clusters:int = None, Sfunc = lambda x:np.mean(x,0) ) -> tuple :
     clabels_n , clabels_o = None , None
     res         = sclinkages( distm , cmd )['F']
     index       = list( res.keys() )
     if labels is None :
         labels = range(len(distm))
     hierarch_df = pd.DataFrame ( res.values() , index=list(res.keys()) , columns = labels )
-    cluster_df  = hierarch_df .T .apply( lambda x: cluster_appraisal(x,garbage_n = 0) )
+    cluster_df  = hierarch_df .T .apply( lambda x: cluster_appraisal(x , garbage_n = 0 , Sfunc=Sfunc ) )
     clabels_o , clabels_n = None , None
     screening    = np.array( [ v[-1][0] for v in cluster_df.values ] )
     level_values = np.array( list(res.keys()) )
@@ -1323,6 +1323,53 @@ def generate_clustering_labels ( distm:np.array , cmd:str='min' , labels:list[st
         clabels_n = hierarch_df.iloc[jhit,:].values.tolist()
     return ( clabels_n , clabels_o , hierarch_df , np.array( [level_values,screening] ) )
 
+def sL( L:list[str] ) -> pd.DataFrame :
+    n = len(L)
+    m = n
+    K = np.meshgrid( range(n),range(m) ) [ 0 ] # WASTEFUL
+    return ( pd.DataFrame( np.array([ L[i] for i in K.reshape(-1) ]).reshape(n,m) ) +'.'+\
+             pd.DataFrame( np.array([ L[i] for i in K.reshape(-1) ]).reshape(n,m) ).T )
+
+def sB( L:list[str], logic=lambda a,b : a==b ) -> list[bool] :
+    convert_labels = lambda T : [ {s:i for s,i in zip(  list(set(T)),range(len(set(T)))  ) }[l] for l in T ]
+    L = convert_labels(L)
+    n = len(L)
+    m = n
+    K = np.meshgrid( range(n),range(m) ) [ 0 ] # WASTEFUL
+    return (  logic( np.array([ L[i] for i in K.reshape(-1) ]).reshape(n,m) ,
+             np.array([ L[i] for i in K.reshape(-1) ]).reshape(n,m).T ) )
+
+def pdB( L:list[str] , logic = lambda a,b:a==b , type_function = lambda x:int(x) ) -> list[int] :
+    return ( pd.DataFrame( sB(L=L,logic=logic) ).apply(lambda x:x.apply(type_function) ) )
+
+def sB12( L1:list[str],L2:list[str] ) -> list[bool] :
+    n = len(L1)
+    m = len(L2)
+    K = np.meshgrid( range(n),range(m) )
+    return ( np.array([ L1[i] for i in K[0].reshape(-1) ]).reshape(n,m) ==\
+             np.array([ L2[i] for i in K[1].reshape(-1) ]).reshape(n,m).T )
+
+def label_correspondances ( L1:list[str] , L2:list[str] ,
+                            bSymmetric:bool = False , bVerbose:bool = False,
+                            bSingleton:bool = False ) -> list[int] :
+    if bVerbose:
+        print ( "EVALUATES GROUPING INTERACTIONS SO SINGLETON LABELING CAN INCLUDED" )
+        print ( "INCLUDE IT BY SETTING bSingleton=True OR False TO EXCLUDE THE DIAGONAL" )
+    nssum = lambda x:np.sum(np.sum(x))
+    if len(L1) != len(L2) :
+        print ( "WARNING : UNEQUAL LENGTHS DOESN'T MAKE SENSE" )
+        return ( [0,0,0,0] )
+    E = 1 - np.eye(len(L1))
+    if bSingleton :
+        E = 1  
+    p_eS1 = pdB( L1 , logic = lambda a,b: a==b ) * E
+    p_eS2 = pdB( L2 , logic = lambda a,b: a==b ) * E
+    TP = nssum( ( p_eS1 ) * ( p_eS2 ) > 0)      # TP
+    FN = nssum( ( 1-p_eS1 ) * (p_eS2) > 0)      # FN
+    FP = nssum( ( p_eS1 ) * (1-p_eS2) > 0)      # FP
+    TN = nssum( ( 1-p_eS1 ) * (1-p_eS2) > 0)    # TN
+    return ( [TP,FP,FN,TN] )
+
 
 if __name__ == '__main__' :
 
