@@ -1,5 +1,5 @@
 """
-Copyright 2021 RICHARD TJÖRNHAMMAR
+Copyright 2023 RICHARD TJÖRNHAMMAR
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -35,11 +35,12 @@ def pathway_frame_from_file ( filename ,
             pdf = pd.concat([pdf,pd.DataFrame(ps).T])
     return ( pdf )
 
-def create_dag_representation_df ( pathway_file = '../data/GROUPDEFINITIONS.gmt' ,
-                                   pcfile = '../data/PCLIST.txt'
+
+def create_dag_representation_df ( pathway_file:str = '../data/GROUPDEFINITIONS.gmt' ,
+                                   pcfile:str = '../data/PCLIST.txt', identifier:str='R-HSA'
                                  ) :
     pc_list_file = pcfile
-    tree , ance , desc = parent_child_to_dag ( pc_list_file )
+    tree , ance , desc = parent_child_to_dag ( pc_list_file , identifier=identifier )
     pdf = make_pathway_ancestor_data_frame ( ance )
     pdf_ = pathway_frame_from_file( pathway_file )
     pdf.index = [v.replace(' ','') for v in  pdf.index.values]
@@ -47,6 +48,7 @@ def create_dag_representation_df ( pathway_file = '../data/GROUPDEFINITIONS.gmt'
     dag_df = pd.concat([pdf.T,pdf_.T]).T
     return ( dag_df , tree )
 
+
 def HierarchalEnrichment (
             analyte_df , dag_df , dag_level_label = 'DAG,l' ,
             ancestors_id_label = 'aid' , id_name = None , threshold = 0.05 ,
@@ -60,6 +62,7 @@ def HierarchalEnrichment (
             p_label = p_label , analyte_name_label = analyte_name_label ,
             item_delimiter = item_delimiter , alexa_elim = alexa_elim , alternative = alternative ) )
 
+
 def HierarchicalEnrichment (
             analyte_df , dag_df , dag_level_label = 'DAG,l' ,
             ancestors_id_label = 'aid' , id_name = None , threshold = 0.05 ,
@@ -115,10 +118,87 @@ def HierarchicalEnrichment (
     df = df.dropna()
     return ( df )
 
+
+
+def hierarchy_matrix ( distance_matrix:np.array   = None ,
+                       coordinates:np.array       = None ,
+                       linkage_distances:np.array = None ) -> dict :
+    from impetuous.clustering import connectivity , absolute_coordinates_to_distance_matrix
+    import operator
+    if not operator.xor( coordinates is None , distance_matrix is None ) :
+        print ( "ONLY COORDINATES OR A DISTANCE MATRIX" )
+        print ( "calculate_hierarchy_matrix FAILED" )
+        print ( "DATA MATRICES NEEDS TO BE SPECIFIED WITH \" distance_matrix = ... \" " )
+        exit(1)
+    if not coordinates is None :
+        distance_matrix = absolute_coordinates_to_distance_matrix(coordinates)
+
+    nmt_ = np.shape(distance_matrix)
+    if linkage_distances is None :
+        uco_v = sorted(list(set(distance_matrix.reshape(-1))))
+    else :
+        uco_v = sorted(list(set(linkage_distances.reshape(-1))))
+
+    level_distance_lookup = {}
+    hsers = []
+    for icut in range(len(uco_v)) :
+        cutoff = uco_v[icut]
+        # clustercontacts : clusterid , particleid relation
+        # clustercontent : clusterid to number of particles in range
+        #from clustering import connectivity # LOCAL TESTING
+        clustercontent , clustercontacts = connectivity ( distance_matrix , cutoff )
+        #
+        # internal ordering is a range so this does not need to be a dict
+        level_distance_lookup[icut] = [ icut , cutoff , np.mean(clustercontent) ]
+        hsers.append(clustercontacts[:,0])
+        if len( set(clustercontacts[:,0]) ) == 1 : # DON'T DO HIGHER COMPLETE SYSTEM VALUES
+            break
+    return ( { 'hierarchy matrix':np.array(hsers) , 'lookup':level_distance_lookup} )
+
+def reformat_hierarchy_matrix_results ( hierarchy_matrix:np.array , lookup:dict=None ) :
+    CL = {}
+    for i in range(len(hierarchy_matrix)):
+        row = hierarchy_matrix[i]
+        d = i
+        if not lookup is None :
+            d   = lookup[i][1]
+        sv_ = sorted(list(set(row)))
+        cl  = {s:[] for s in sv_}
+        for i in range( len( row ) ) :
+            cl[row[i]].append(i)
+        for v_ in list( cl.values() ) :
+            if tuple(v_) not in CL :
+                CL[ tuple(v_) ] = d
+    return ( CL )
+
+def build_snn_hierarchy ( distm:np.array ):
+    #
+    # print ( "BUILDING SYMMETRY BROKEN HIERARCHY" )
+    if True :
+        nnvec                   = []
+        hierarchy_matrix        = []
+        N                       = np.min( np.shape(distm) )
+        from impetuous.clustering import nearest_neighbor_graph_matrix , connectivity_boolean
+        for NN in range ( 1 , N ) :
+            nnvec .append( NN )
+            NN_bool     = nearest_neighbor_graph_matrix ( distm , NN )[0]<np.inf
+            solution    = connectivity_boolean ( NN_bool )
+            hierarchy_matrix .append( [ cid[0] for cid in solution[1] ] )
+            if len(solution[0])>1 :
+                continue
+            else :
+                break
+    else :
+        print ( "DONE AT" , N )
+    return ( hierarchy_matrix )
+
 def calculate_hierarchy_matrix ( data_frame = None ,
                                  distance_matrix = None ,
                                  bVerbose = False,
                                  coarse_grain_structure = 0 ) :
+
+    print ( "WARNING:LEGACY METHOD. CHECK IF hierarchy_matrix MIGHT SUIT YOU BETTER" )
+
     info__ = """ This is the saiga/pelican/panda you are looking for RICEARD"""
     # print (info__ )
     from impetuous.clustering import connectivity , absolute_coordinates_to_distance_matrix
@@ -326,38 +406,66 @@ def write_cpgmt ( lookup ,
                                  (c,d,p) in zip( *[ lookup[ c ] for c in lookup['content'] ]) ]) ,
                     file=of )
 
+#
+# TOOLS LOCAL
 def ordered_remove ( str,delete ):
     for d in delete :
         str = str.replace(d,'')
     return ( str )
 
-
-def build_pclist_word_hierarchy ( filename = 'new_compartment_genes.gmt',
-    delete   = ['\n'] , group_id_prefix = 'COMP',
-    analyte_prefix  = 'ENSG', root_name = 'COMP0000000000',
-    bReturnList = False , bUseGroupPrefix = False, bSingleChild=True ):
-
-    if not '.gmt' in filename :
-        print ( 'MUST HAVE A VALID GMT FILE' )
+def error ( criterion,message ) :
+    if criterion :
+        print ( message )
         exit ( 1 )
 
-    # RETURNS THE PC LIST THAT CREATES THE WORD HIERARCHY
-    # LATANTLY PRESENT IN THE GMT ANALYTE DEFINITIONS
+def build_pclist_word_hierarchy ( filename = None ,  # 'new_compartment_genes.gmt',
+                                  ledger   = None ,
+    delete   = ['\n'] , group_id_prefix = None,
+    analyte_prefix  = 'ENSG', root_name = 'COMP0000000000',
+    bReturnList = False , bUseGroupPrefix = False ,
+    bSingleChild = False , bSingleDescent = True ) :
 
+    bSingleChild = bSingleChild or bSingleDescent # USER SHOULD SET THE bSingleDescent OPTION
+
+    bUseFile = not filename is None
+    import operator
+    error ( not operator.xor ( filename is None , ledger is None ), "YOU MUST SUPPLY A GMT FILE XOR A DICTIONARY" )
+    if bUseFile :
+        error ( not '.gmt' in filename , 'MUST HAVE A VALID GMT FILE' )
+    #
+    # RETURNS THE PC LIST THAT CREATES THE WORD HIERARCHY
+    # LATANTLY PRESENT IN THE GMT ANALYTE (GROUPING) DEFINITIONS
+    #
     S_M = set()
     D_i = dict()
 
+    bUseGroupPrefix = not group_id_prefix is None
+    if bUseGroupPrefix :
+        bUseGroupPrefix = 'str' in str(type(group_id_prefix)).lower()
     check_prefix = analyte_prefix
     if bUseGroupPrefix :
         check_prefix = group_id_prefix
-    
-    with open ( filename,'r' ) as input :
-        for line in input :
-            lsp = ordered_remove(line,delete).split('\t')
-            if not check_prefix in line :
-                continue
-            S_i = set(lsp[2:])
-            D_i[ lsp[0] ] = tuple( (lsp[1],S_i,len(S_i)) )
+
+    if bUseFile :
+        with open ( filename,'r' ) as input :
+            for line in input :
+                lsp = ordered_remove(line,delete).split('\t')
+                if not check_prefix in line :
+                    continue
+                S_i = set(lsp[2:])
+                D_i [ lsp[0] ] = tuple( (lsp[1] , S_i , len(S_i)) )
+                S_M = S_M | S_i
+    else :
+        for item in ledger.items() :
+            print(item)
+            if bUseGroupPrefix :
+                if not check_prefix in item[0]:
+                    continue
+            else :
+                if not check_prefix in ''.join(item[1][1]):
+                    continue
+            S_i = set( item[1][1] )
+            D_i [ item[0] ] = tuple( (item[1][0] , S_i , len(S_i)) )
             S_M = S_M | S_i
 
     isDecendant  = lambda sj,sk : len(sj-sk)==0
@@ -417,6 +525,21 @@ def build_pclist_word_hierarchy ( filename = 'new_compartment_genes.gmt',
     else :
         return ( PClist )
 
+def matrixZ2linkage_dict_tuples ( Z :np.array ) -> dict :
+    # from scipy.cluster.hierarchy import linkage
+    from scipy.cluster.hierarchy import fcluster
+    CL = {}
+    for d in Z[:,2] :
+        row = fcluster ( Z ,d, 'distance' )
+        sv_ = sorted(list(set(row)))
+        cl  = {s:[] for s in sv_}
+        for i in range( len( row ) ) :
+            cl[row[i]].append(i)
+        for v_ in list( cl.values() ) :
+            CL[tuple(v_)] = d
+    return ( CL )
+
+
 if __name__ == '__main__' :
 
     if False :
